# pylint: disable=unused-import
from identities.serializers.permission import (
    AnyObjectSerializer,
    AnyObjectByPermissionSerializer,
    PermissionSerializer,
)
from identities.serializers.group import GroupSerializer
from identities.serializers.user import UserSerializer
from identities.serializers.userinfo import UserinfoSerializer



# Generated by Django 3.2.4 on 2021-08-02 07:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("identities", "0003_historicaluser"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicaluser",
            name="id",
            field=models.BigIntegerField(
                auto_created=True, blank=True, db_index=True, verbose_name="ID"
            ),
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("identities", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalprofile",
            name="is_tech_account",
        ),
        migrations.RemoveField(
            model_name="profile",
            name="is_tech_account",
        ),
    ]



from django.urls import re_path
from django.conf.urls import include
from rest_framework.routers import DefaultRouter

from . import views, oidc
from .apps import APP_NAME


router = DefaultRouter()
router.register(r"users", views.UserViewSet, "user")
router.register(r"permission", views.PermissionViewSet, "permission")
router.register(
    r"object_by_permission", views.ObjectByPermissionViewSet, "object_by_permission"
)
router.register(r"group", views.GroupViewSet, "group")


# The API URLs are now determined automatically by the router.
app_name = APP_NAME
urlpatterns = [
    re_path(r"", include(router.urls)),
    re_path("^auth/oidc/authenticate/", oidc.authenticate, name="oidc-authenticate"),
    re_path("^auth/oidc/callback/", oidc.callback, name="oidc-callback"),
    re_path("^auth/oidc/logout/", oidc.logout, name="oidc-logout"),
    re_path(r"^userinfo/$", views.UserinfoView.as_view(), name="userinfo"),
]



import logging
from typing import Optional

from django.contrib.auth.models import AbstractUser, Group
from django.db.models import QuerySet
from guardian.shortcuts import get_objects_for_user
from rest_framework import permissions
from rest_framework.exceptions import MethodNotAllowed
from rest_framework.permissions import BasePermission, IsAuthenticated

logger = logging.getLogger(__name__)
update_methods = ("PUT", "PATCH")
perm_group_manager = "auth.change_group"


def get_group_manager_groups(user: AbstractUser) -> QuerySet[Group]:
    return get_objects_for_user(user, perm_group_manager)


def has_group_manager_groups(user: AbstractUser) -> bool:
    return get_group_manager_groups(user).exists()


def has_group_manager_permission(
    user: AbstractUser, group: Optional[Group] = None
) -> bool:
    return user.has_perm(perm_group_manager, group)


class IsGroupManagerRead(permissions.BasePermission):
    """Group Manager can modify a Group"""

    def has_permission(self, request, view):
        user = request.user
        return (
            user
            and user.is_authenticated
            and has_group_manager_groups(user)
            and request.method in ("GET", "HEAD")
        )

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view) and has_group_manager_permission(
            request.user, obj
        )


class IsGroupManagerWrite(permissions.BasePermission):
    """Group Manager can edit a Group"""

    def has_permission(self, request, view):
        user = request.user
        return (
            user
            and user.is_authenticated
            and has_group_manager_groups(user)
            and request.method in ("PUT", "PATCH")
        )

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view) and has_group_manager_permission(
            request.user, obj
        )


def is_staff(user: AbstractUser) -> bool:
    return user.is_staff or user.is_superuser


class AuditPermission(permissions.BasePermission):
    """Staff can read"""

    def has_permission(self, request, view):
        return (
            request.method in permissions.SAFE_METHODS
            and request.user
            and request.user.is_authenticated
            and request.user.is_staff
        )


class IsStaff(IsAuthenticated):
    """Staff can read/write everything."""

    def has_permission(self, request, view):
        return super().has_permission(request, view) and request.user.is_staff

    def has_object_permission(self, request, view, _):
        """Can access and perform any action on any object"""
        return self.has_permission(request, view)


class ReadOnly(IsAuthenticated):
    """Authorized users with read-only access."""

    def has_permission(self, request, view):
        return (
            super().has_permission(request, view)
            and request.method in permissions.SAFE_METHODS
        )

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)


class IsStaffOrReadOnly(permissions.BasePermission):
    """Authenticated users can view, staff can edit."""

    def has_permission(self, request, view):
        return (
            request.user
            and request.user.is_authenticated
            and (request.method in permissions.SAFE_METHODS or request.user.is_staff)
        )


class IsAuthenticatedAndUniqueCheck(permissions.BasePermission):
    """Allows authenticated users to perform a unique check"""

    def has_permission(self, request, view):
        return (
            request.user and request.user.is_authenticated and request.method == "POST"
        )


class UserUniquePermission(permissions.BasePermission):
    methods = ("POST",)

    def has_permission(self, request, view):
        return (
            request.user
            and request.user.is_authenticated
            and request.method in self.methods
            and is_staff(request.user)
        )


class UserDataPermission(permissions.BasePermission):
    update_user_method = "PATCH"
    create_local_user_method = "POST"

    def has_permission(self, request, view):
        if request.user and request.user.is_authenticated:
            if request.method == self.create_local_user_method:
                return is_staff(request.user)
            return True
        return False

    # pylint: disable=too-many-return-statements
    def has_object_permission(self, request, view, obj):  # NOT called on POST!
        if request.method in permissions.SAFE_METHODS:
            # non-modifying operation
            return True

        # modifying operation
        requestor = request.user

        if request.method == self.update_user_method:
            # attempt to modify user
            if not is_staff(requestor):
                logger.error(
                    "Non-staff user (ID: %s) attempted to modify user (ID: %s) with data: %s",
                    requestor.id,
                    obj.id,
                    request.data,
                )
                return False

            return True

        # not allowed modifying method
        logger.error(
            "User (ID: %s) attempted to modify user (ID: %s)"
            "using not allowed method %s with data: %s",
            requestor.id,
            obj.id,
            request.method,
            request.data,
        )
        return False


class ObjectPermission(BasePermission):
    """Check if user has correct object-level permission for the action

    This class is roughly based on rest_framework.permissions.DjangoObjectPermissions
    """

    perms_map = {
        "GET": ["{app_label}.view_{model_name}"],
        "OPTIONS": [],
        "HEAD": ["{app_label}.view_{model_name}"],
        "PUT": ["{app_label}.change_{model_name}"],
        "PATCH": ["{app_label}.change_{model_name}"],
        "DELETE": ["{app_label}.delete_{model_name}"],
    }

    def get_required_object_permissions(self, method, model_cls):
        kwargs = {
            "app_label": model_cls._meta.app_label,
            "model_name": model_cls._meta.model_name,
        }

        if method not in self.perms_map:
            raise MethodNotAllowed(method)

        return tuple(perm.format(**kwargs) for perm in self.perms_map[method])

    def has_permission(self, request, view):
        if not request.user or not request.user.is_authenticated:
            return False
        if request.method not in self.perms_map:
            raise MethodNotAllowed(request.method)
        return True

    def has_object_permission(self, request, view, obj):
        perms = self.get_required_object_permissions(request.method, obj.__class__)
        return request.user.has_perms(perms, obj)



import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser

from identities.user_utils import create_user, update_user


User = get_user_model()

CREATION_CLAIMS = {
    "preferred_username": "906726174130@identity.ch",
    "given_name": "Bob",
    "family_name": "Smith",
    "email": "bob@example.org",
    "linkedAffiliation": ["member@sciences.ch", "staff@sciences.ch"],
    "linkedAffiliationUniqueID": ["987321@sciences.ch", "24680@university.ch"],
}


def create_bob(
    affiliation="member@hospital.ch,staff@hospital.ch",
    affiliation_id="1234567@zuerich.ch,7654321@university.ch",
) -> AbstractUser:
    user = User.objects.create_user(username="bob", email="bob@example.org")
    user.profile.affiliation = affiliation
    user.profile.affiliation_id = affiliation_id
    return user


@pytest.mark.django_db
def test_create_user():
    user_created = create_user(CREATION_CLAIMS)
    user_db = User.objects.get(username="906726174130@identity.ch")
    assert user_created == user_db
    assert user_created.username == "906726174130@identity.ch"
    assert user_created.first_name == "Bob"
    assert user_created.last_name == "Smith"
    assert user_created.email == "bob@example.org"
    assert user_created.profile.affiliation == "member@sciences.ch,staff@sciences.ch"
    assert (
        user_created.profile.affiliation_id == "987321@sciences.ch,24680@university.ch"
    )
    assert create_user({}) is None


@pytest.mark.django_db
def test_from_local_to_federated():
    user_db = create_bob()
    assert user_db.username == "bob"
    assert user_db.first_name == ""
    assert user_db.last_name == ""
    assert user_db.profile.affiliation == "member@hospital.ch,staff@hospital.ch"
    assert user_db.profile.affiliation_id == "1234567@zuerich.ch,7654321@university.ch"
    user_updated = create_user(CREATION_CLAIMS)
    assert user_updated.username == "906726174130@identity.ch"
    assert user_updated.first_name == "Bob"
    assert user_updated.last_name == "Smith"
    assert user_updated.profile.affiliation == "member@sciences.ch,staff@sciences.ch"
    assert (
        user_updated.profile.affiliation_id == "987321@sciences.ch,24680@university.ch"
    )


@pytest.mark.django_db
def test_federation_overwrites_local():
    user_db = create_bob()
    user_db.first_name = "Robert"
    user_db.last_name = "Martin"
    user_db.save()
    user_updated = create_user(CREATION_CLAIMS)
    assert user_updated.username == "906726174130@identity.ch"
    assert user_updated.first_name == "Bob"
    assert user_updated.last_name == "Smith"


@pytest.mark.django_db
def test_update_user_notification():
    user_db = create_bob()
    old_affiliation = user_db.profile.affiliation
    old_affiliation_id = user_db.profile.affiliation_id
    assert old_affiliation == "member@hospital.ch,staff@hospital.ch"
    assert old_affiliation_id == "1234567@zuerich.ch,7654321@university.ch"
    user_updated = update_user(
        user_db,
        {k: v for k, v in CREATION_CLAIMS.items() if k != "preferred_username"},
    )
    assert User.objects.get(username="bob") == user_updated
    assert user_updated.username == "bob"
    assert user_updated.first_name == "Bob"
    assert user_updated.last_name == "Smith"
    assert user_updated.email == "bob@example.org"
    # Affiliation has changed
    new_affiliation = "member@sciences.ch,staff@sciences.ch"
    new_affiliation_id = "987321@sciences.ch,24680@university.ch"
    assert user_updated.profile.affiliation == new_affiliation
    assert user_db.profile.affiliation == new_affiliation
    assert user_updated.profile.affiliation_id == new_affiliation_id
    assert user_db.profile.affiliation_id == new_affiliation_id


# pylint: disable=unused-argument
@pytest.mark.parametrize(
    "old_affiliation,new_affiliation,called",
    (
        ("", "", False),
        (None, "", False),
        ("", None, False),
        ("", "staff@sciences.ch", True),
        ("staff@sciences.ch", None, True),
        ("member@sciences.ch", "member@sciences.ch", False),
        (
            "staff@sciences.ch,member@sciences.ch",
            "staff@sciences.ch,member@sciences.ch",
            False,
        ),
        (
            "member@sciences.ch,staff@sciences.ch",
            "staff@sciences.ch,member@sciences.ch",
            False,
        ),
        ("student@sciences.ch,staff@sciences.ch", "staff@sciences.ch", False),
        (
            "student@sciences.ch,staff@sciences.ch",
            "staff@sciences.ch,member@sciences.ch",
            False,
        ),
        (
            "student@zuerich.ch,staff@sciences.ch",
            "student@sciences.ch,member@zuerich.ch",
            False,
        ),
        ("student@sciences.ch,staff@sciences.ch", "student@zuerich.ch", True),
        ("student@sciences.ch,student@zuerich.ch", "student@zuerich.ch", True),
    ),
)
@pytest.mark.django_db
def test_update_user(old_affiliation, new_affiliation, called):
    user_db = create_bob(old_affiliation)
    assert user_db.profile.affiliation == old_affiliation

    def empty_if_none(value):
        return value if value is not None else ""

    user_updated = update_user(
        user_db,
        {
            "linkedAffiliation": new_affiliation.split(",") if new_affiliation else [],
        },
    )
    # Affiliation has changed
    assert user_updated.profile.affiliation == empty_if_none(new_affiliation)
    assert user_db.profile.affiliation == empty_if_none(new_affiliation)



from django.contrib.auth import get_user_model
from guardian.shortcuts import get_objects_for_user
from rest_framework.filters import BaseFilterBackend

User = get_user_model()


class UserActiveFilter(BaseFilterBackend):
    """Filter users depending on the user being active (`is_active` is `True`)."""

    param = "include_inactive"

    def filter_queryset(self, request, queryset, view):
        if request.method != "GET" or request.query_params.get(self.param):
            # pylint: disable=no-member
            return queryset.all()
        # exclude inactive users by default
        return queryset.exclude(is_active=False)

    def get_schema_operation_parameters(self, view):
        return (
            {
                "name": self.param,
                "required": False,
                "in": "query",
                "description": "Includes users which are not active (`is_active` is `False`)",
                "schema": {"type": "boolean"},
            },
        )


class ObjectPermissionsFilter(BaseFilterBackend):
    """Filter objects which user has object-level view permission"""

    accept_global_permissions = False
    superuser_show_all = True

    def filter_queryset(self, request, queryset, view):
        app_label = queryset.model._meta.app_label
        model_name = queryset.model._meta.model_name
        obj = get_objects_for_user(
            request.user,
            f"{app_label}.view_{model_name}",
            queryset,
            accept_global_perms=self.accept_global_permissions,
            with_superuser=self.superuser_show_all,
        )
        return obj



import logging

from authlib.integrations.base_client import MismatchingStateError
from authlib.integrations.django_client import OAuth
from django.conf import settings
from django.contrib import auth
from django.shortcuts import redirect
from django.urls import reverse

from .apps import APP_NAME
from .user_utils import user_from_claims

logger = logging.getLogger(__name__)

oauth = OAuth()
oauth.register(name="identity_provider")


def authenticate(request):
    redirect_uri = request.build_absolute_uri(reverse(f"{APP_NAME}:oidc-callback"))
    return oauth.identity_provider.authorize_redirect(request, redirect_uri)


def callback(request):
    try:
        token = oauth.identity_provider.authorize_access_token(request)
        userinfo = oauth.identity_provider.userinfo(token=token)
        user = user_from_claims(username_claim="preferred_username", claims=userinfo)
        if user is not None:
            auth.login(
                request, user, backend="django.contrib.auth.backends.ModelBackend"
            )
        return redirect(settings.LOGIN_REDIRECT_URL)
    except MismatchingStateError as e:
        logger.warning(
            "User likely tried to login using a URL including an expired session (in a bookmark). "
            "Details: %s",
            e,
            exc_info=True,
        )
        return redirect(settings.LOGIN_REDIRECT_URL)


def logout(request):
    auth.logout(request)
    provider_logout_endpoint = oauth.identity_provider.server_metadata.get(
        "end_session_endpoint"
    )
    return redirect(provider_logout_endpoint or settings.LOGOUT_REDIRECT_URL)



import logging

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db import transaction
from django_drf_utils.serializers.utils import get_request_username
from rest_framework import serializers
from rest_framework.exceptions import MethodNotAllowed

from ..serializers.profile import ProfileSerializer


logger = logging.getLogger(__name__)
User = get_user_model()


class UserSerializer(serializers.ModelSerializer):
    profile = ProfileSerializer(required=False)
    groups = serializers.SlugRelatedField(
        slug_field="name",
        read_only=False,
        many=True,
        required=False,
        queryset=Group.objects.all(),
    )

    class Meta:
        model = User
        read_only_fields = (
            "id",
            "username",
            "last_login",
        )
        fields = read_only_fields + (
            "profile",
            "email",
            "first_name",
            "last_name",
            "groups",
            "is_active",
        )

    @transaction.atomic
    def update(self, instance, validated_data):
        if "groups" in validated_data:
            instance.groups.set(validated_data["groups"])
        if (
            "is_active" in validated_data
            and validated_data["is_active"] != instance.is_active
        ):
            request_user = get_request_username(self)
            logger.info(
                "User (ID: %s) changed `is_active` of user (ID: %s) from %s to %s",
                request_user.id,
                instance.id,
                instance.is_active,
                validated_data["is_active"],
            )
            instance.is_active = validated_data["is_active"]

            if not instance.is_active:
                # remove all associations to user
                # remove all permissions
                instance.groups.clear()
                instance.user_permissions.clear()

        instance.save()
        return instance

    def create(self, validated_data):
        if "profile" in validated_data:
            del validated_data["profile"]
        validated_data["username"] = validated_data["email"]
        return super().create(validated_data)


class UserShortSerializer(serializers.ModelSerializer):
    """A read-only user serializer."""

    affiliation = serializers.CharField(source="profile.affiliation", read_only=True)
    affiliation_id = serializers.CharField(
        source="profile.affiliation_id", read_only=True
    )

    class Meta:
        model = User
        read_only_fields = (
            "username",
            "email",
            "last_name",
            "first_name",
            "affiliation",
            "affiliation_id",
        )
        fields = read_only_fields

    # Make sure that this serializer is NOT used for POST or PUT requests

    def create(self, validated_data):
        raise MethodNotAllowed(method="POST")

    def update(self, instance, validated_data):
        raise MethodNotAllowed(method="PUT")



# Generated by Django 3.2.4 on 2021-07-28 15:11

from django.conf import settings
import django.contrib.auth.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import simple_history.models


class Migration(migrations.Migration):
    dependencies = [
        ("identities", "0002_remove_tech_account"),
    ]

    operations = [
        migrations.CreateModel(
            name="HistoricalUser",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        db_index=True,
                        error_messages={
                            "unique": "A user with that username already exists."
                        },
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        validators=[
                            django.contrib.auth.validators.UnicodeUsernameValidator()
                        ],
                        verbose_name="username",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="first name"
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="last name"
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        blank=True,
                        db_index=True,
                        max_length=254,
                        verbose_name="email address",
                    ),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical user",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



from django.contrib.auth import get_user_model
from django_drf_utils.views.utils import unique_check
from rest_framework import mixins, permissions, viewsets, generics, serializers
from rest_framework.filters import SearchFilter
from rest_framework.schemas.openapi import AutoSchema

from ..filters import UserActiveFilter
from ..permissions import (
    has_group_manager_groups,
    UserDataPermission,
    UserUniquePermission,
)
from ..serializers import (
    UserSerializer,
    UserinfoSerializer,
)

User = get_user_model()


class UserinfoSchema(AutoSchema):
    def map_field(self, field):
        mapped_field = super().map_field(field)
        if isinstance(field, serializers.SerializerMethodField):
            field_name_to_component = {
                "permissions": {
                    "type": "object",
                    "readOnly": True,
                    "properties": {
                        "staff": {"readOnly": True, "type": "boolean"},
                        "group_manager": {"readOnly": True, "type": "boolean"},
                    },
                },
                "manages": {
                    "readOnly": True,
                    "type": "object",
                    "properties": {
                        "data_providers": {
                            "readOnly": True,
                            "type": "array",
                            "items": {"$ref": "#/components/schemas/DataProvider"},
                        }
                    },
                },
            }
            return field_name_to_component.get(field.field_name, mapped_field)
        return mapped_field


# pylint: disable=too-many-ancestors
@unique_check((UserUniquePermission,))
class UserViewSet(
    mixins.ListModelMixin,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    viewsets.GenericViewSet,
):
    """Show details of registered users."""

    serializer_class = UserSerializer
    permission_classes = (UserDataPermission,)
    filter_backends = (SearchFilter, UserActiveFilter)
    search_fields = ("username", "email")

    def get_queryset(self):
        if self.request.user.is_staff or has_group_manager_groups(self.request.user):
            return User.objects.all()
        # Restrict unnecessary access to other people data
        return User.objects.filter(username=self.request.user.username)

    def create(self, request, *args, **kwargs):
        # False positive, exists on mixins.CreateModelMixin
        # pylint: disable=no-member
        response = super().create(request, *args, **kwargs)
        created_user = User.objects.get(id=response.data["id"])
        created_user.set_unusable_password()
        created_user.save()
        return response


class UserinfoView(generics.RetrieveAPIView):
    serializer_class = UserinfoSerializer
    permission_classes = (permissions.IsAuthenticated,)
    schema = UserinfoSchema()

    def get_queryset(self):
        return User.objects.filter(username=self.request.user.username)

    def get_object(self):
        return self.get_queryset()[0]



from django.contrib.auth.models import Permission
from rest_framework import serializers


class PermissionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Permission
        fields = "__all__"


# It can serialize an instance of any django.db.models.Model
class AnyObjectSerializer(serializers.Serializer):  # pylint: disable=abstract-method
    id = serializers.IntegerField(read_only=True, min_value=1)
    name = serializers.SerializerMethodField()

    def get_name(self, obj):
        return str(obj)


# It can serialize an instance of any django.db.models.Model
class AnyObjectByPermissionSerializer(
    serializers.Serializer
):  # pylint: disable=abstract-method
    perm_id = serializers.IntegerField(read_only=True, min_value=1)
    obj_id = serializers.IntegerField(read_only=True, min_value=1)
    name = serializers.CharField(read_only=True)



import logging

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.db.transaction import atomic
from django_drf_utils.serializers.utils import get_request_username
from guardian.models import GroupObjectPermission
from rest_framework import serializers

from ..permissions import is_staff


logger = logging.getLogger(__name__)
User = get_user_model()

VALIDATION_ERROR_NON_STAFF_FIELDS = (
    "Only staff are allowed to change fields other than `users`"
)


class GroupObjectPermissionSerializer(
    serializers.Serializer
):  # pylint: disable=abstract-method
    permission = serializers.IntegerField()
    objects = serializers.ListField(child=serializers.IntegerField(min_value=1))

    def validate_permission(self, value):
        try:
            return Permission.objects.get(pk=value)
        except Permission.DoesNotExist as e:
            raise serializers.ValidationError(
                f"Permission {value} does not exist"
            ) from e

    def validate(self, attrs):
        data = super().validate(attrs)
        model = data["permission"].content_type.model_class()
        for obj in data["objects"]:
            try:
                model.objects.get(pk=obj)
            except model.DoesNotExist as e:
                raise serializers.ValidationError(
                    f"Object {obj} does not exist for model {model.__name__}"
                ) from e
        return data


class GroupObjectPermissionListSerializer(
    serializers.ListSerializer
):  # pylint: disable=abstract-method
    def to_representation(self, data):
        output = {}
        for perm in data.all():
            output.setdefault(perm.permission.id, []).append(perm.object_pk)
        return [
            GroupObjectPermissionSerializer({"permission": p, "objects": o}).data
            for p, o in output.items()
        ]


class GroupSerializer(serializers.ModelSerializer):
    permissions_object = GroupObjectPermissionListSerializer(
        child=GroupObjectPermissionSerializer(),
        required=False,
        source="groupobjectpermission_set",
    )
    users = serializers.PrimaryKeyRelatedField(
        many=True, source="user_set", queryset=User.objects.all()
    )

    class Meta:
        model = Group
        fields = ("id", "name", "users", "permissions", "permissions_object")

    def _log_action(self, action: str, queryset):
        if queryset:
            logger.info("%s %s %s", get_request_username(self), action, queryset)

    @staticmethod
    def _transform_object_permissions(permissions):
        return (
            (perm_grp["permission"].content_type, str(obj), perm_grp["permission"])
            for perm_grp in permissions
            for obj in perm_grp["objects"]
        )

    def _create_object_permissions(self, obj_perm, instance: Group):
        created = GroupObjectPermission.objects.bulk_create(
            GroupObjectPermission(
                group=instance, content_type=t, object_pk=o, permission=p
            )
            for t, o, p in obj_perm
        )
        self._log_action("created", created)

    @atomic
    def create(self, validated_data):
        obj_perm = self._transform_object_permissions(
            validated_data.pop("groupobjectpermission_set", ())
        )
        group = super().create(validated_data)
        self._create_object_permissions(obj_perm, group)
        return group

    def _validate_non_staff_changes(self, instance, validated_data, new_perm, old_perm):
        user = get_request_username(self)
        current_permissions = list(instance.permissions.all())
        has_different_name = validated_data.get("name", instance.name) != instance.name
        has_different_permissions = (
            validated_data.get("permissions", current_permissions)
            != current_permissions
        )
        has_different_group_object_permissions = set(old_perm.keys()) != new_perm
        if not is_staff(user) and (
            has_different_name
            or has_different_permissions
            or has_different_group_object_permissions
        ):
            raise serializers.ValidationError(VALIDATION_ERROR_NON_STAFF_FIELDS)

    @atomic
    def update(self, instance, validated_data):
        new_perm = set(
            self._transform_object_permissions(
                validated_data.pop("groupobjectpermission_set", ())
            )
        )
        old_perm = {
            (p.content_type, p.object_pk, p.permission): p.pk
            for p in GroupObjectPermission.objects.filter(group=instance)
        }

        self._validate_non_staff_changes(instance, validated_data, new_perm, old_perm)

        self._create_object_permissions(new_perm - set(old_perm.keys()), instance)
        to_be_removed = GroupObjectPermission.objects.filter(
            pk__in=[
                v for k, v in old_perm.items() if k in set(old_perm.keys()) - new_perm
            ]
        )
        self._log_action("deleted", list(to_be_removed))
        to_be_removed.delete()
        return super().update(instance, validated_data)



import logging
from typing import Any, Dict, Tuple

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.core.exceptions import SuspiciousOperation
from django.conf import settings

User = get_user_model()

logger = logging.getLogger(__name__)


def get_or_create_user(username: str, email: str) -> Tuple[AbstractUser, bool]:
    """Fetches an user or creates a new one.

    Lookup happens via email address (case insensitive).

    :param username: User name. Always specified.
    :param email: email address. Always specified but could be empty.
    :return: tuple of User, bool
    """
    if email:
        try:
            return User.objects.get(email__iexact=email), False
        except User.DoesNotExist:
            # In case of multiple entries (which should hardly happens),
            # exception is passed to caller
            pass
    return User.objects.create_user(username, email), True


def create_user(claims):
    """Return object for a newly created user account."""
    mapper = settings.CONFIG.oidc.mapper
    username = claims.get("preferred_username")
    if not username:
        return None
    email = claims.get(mapper.email, "")
    user, created = get_or_create_user(username, email)
    if not created:
        logger.info(
            "Local user with email '%s' found. Changing username from '%s' to '%s'.",
            email,
            user.username,
            username,
        )
        # Reset the password (federated users have 'None' password) and change username
        user.set_unusable_password()
        user.username = username
    user.first_name = claims.get(mapper.first_name, user.first_name)
    user.last_name = claims.get(mapper.last_name, user.last_name)
    user.profile.affiliation = ",".join(claims.get(mapper.affiliation, []))
    user.profile.affiliation_id = ",".join(claims.get(mapper.affiliation_id, []))
    user.save()
    return user


def update_user(user, claims):
    """Update existing user with new claims, if necessary, save, and return user"""
    mapper = settings.CONFIG.oidc.mapper
    user.first_name = claims.get(mapper.first_name, "")
    user.last_name = claims.get(mapper.last_name, "")
    user.email = claims.get(mapper.email, "")
    sep = ","
    new_affiliation = sep.join(claims.get(mapper.affiliation, []))
    new_affiliation_id = sep.join(claims.get(mapper.affiliation_id, []))
    user.profile.affiliation = new_affiliation
    user.profile.affiliation_id = new_affiliation_id
    user.save()
    return user


def user_from_claims(username_claim: str, claims: Dict[str, Any]):
    """Create or get user based on claims"""
    users = User.objects.filter(username=claims[username_claim])
    if len(users) == 1:
        return update_user(users[0], claims)
    if len(users) == 0:
        return create_user(claims)
    raise SuspiciousOperation("Multiple users returned")



from unittest.mock import Mock

import pytest
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory
from rest_framework import permissions
from rest_framework.exceptions import MethodNotAllowed
from rest_framework.test import APIRequestFactory

import identities.permissions as identities_permission
from identities.apps import APP_NAME
from identities.models import Profile
from identities.permissions import (
    get_group_manager_groups,
    has_group_manager_permission,
    has_group_manager_groups,
    IsGroupManagerRead,
    IsGroupManagerWrite,
    ObjectPermission,
)
from identities.tests.factories import GroupFactory
from tests.conftest import make_group_manager

NONSAFE_METHODS = ("POST", "PUT", "PATCH", "DELETE")
factory = APIRequestFactory()


class TestGroupManager:
    @pytest.fixture(autouse=True)
    def before_each(self, create_group_manager, user_factory):
        # pylint: disable=attribute-defined-outside-init
        self.managed_group = GroupFactory.create()
        self.group_manager = create_group_manager(self.managed_group)
        self.staff = user_factory(staff=True)
        # pylint: enable=attribute-defined-outside-init

    def test_get_group_manager_groups(self):
        assert get_group_manager_groups(self.group_manager)[0] == self.managed_group
        managed_group2 = GroupFactory.create()
        make_group_manager(self.group_manager, managed_group2)
        assert tuple(get_group_manager_groups(self.group_manager)) == (
            self.managed_group,
            managed_group2,
        )
        assert get_group_manager_groups(self.staff).count() == 0

    def test_has_group_manager_groups(self):
        assert has_group_manager_groups(self.group_manager) is True
        assert has_group_manager_groups(self.staff) is False
        make_group_manager(self.staff, self.managed_group)
        assert has_group_manager_groups(self.staff) is True

    def test_has_group_manager_permission(self):
        # only `group_manager` has group manager permission for `managed_group`
        assert (
            has_group_manager_permission(self.group_manager, self.managed_group) is True
        )
        assert has_group_manager_permission(self.staff, self.managed_group) is False
        # `group_manager` does NOT have group manager permission for all groups
        assert has_group_manager_permission(self.group_manager) is False


@pytest.mark.parametrize(
    "permission_obj, allowed_methods, forbidden_methods",
    (
        (IsGroupManagerRead(), ("GET", "HEAD"), ("PUT", "PATCH", "POST", "DELETE")),
        (IsGroupManagerWrite(), ("PUT", "PATCH"), ("GET", "HEAD", "POST", "DELETE")),
    ),
)
class TestIsGroupManager:
    @pytest.fixture(autouse=True)
    def before_each(self, create_group_manager, user_factory):
        # pylint: disable=attribute-defined-outside-init
        self.factory = RequestFactory()
        self.user = user_factory(basic=True)
        self.staff = user_factory(staff=True)
        self.managed_group_1, self.managed_group_2 = GroupFactory.create_batch(2)
        self.group_manager_1 = create_group_manager(self.managed_group_1)
        self.group_manager_2 = create_group_manager(self.managed_group_2)
        self.request = self.factory.request()
        # pylint: enable=attribute-defined-outside-init

    # pylint: disable=unused-argument
    def test_anonymous(self, permission_obj, allowed_methods, forbidden_methods):
        for method in allowed_methods:
            self.request.user = AnonymousUser()
            self.request.method = method
            assert permission_obj.has_permission(self.request, None) is False

    def test_user(self, permission_obj, allowed_methods, forbidden_methods):
        for method in allowed_methods:
            self.request.user = self.user
            self.request.method = method
            assert permission_obj.has_permission(self.request, None) is False

    def test_group_manager_own_group(
        self, permission_obj, allowed_methods, forbidden_methods
    ):
        for user, group in (
            (self.group_manager_1, self.managed_group_1),
            (self.group_manager_2, self.managed_group_2),
        ):
            for method in allowed_methods:
                self.request.user = user
                self.request.method = method
                assert permission_obj.has_permission(self.request, None) is True
                assert (
                    permission_obj.has_object_permission(self.request, None, group)
                    is True
                )

    def test_group_manager_other_group(
        self, permission_obj, allowed_methods, forbidden_methods
    ):
        for user, group in (
            (self.group_manager_1, self.managed_group_2),
            (self.group_manager_2, self.managed_group_1),
        ):
            for method in allowed_methods:
                self.request.user = user
                self.request.method = method
                assert permission_obj.has_permission(self.request, None) is True
                assert (
                    permission_obj.has_object_permission(self.request, None, group)
                    is False
                )

    def test_group_manager_forbidden_methods(
        self, permission_obj, allowed_methods, forbidden_methods
    ):
        for method in forbidden_methods:
            self.request.user = self.group_manager_1
            self.request.method = method
            assert permission_obj.has_permission(self.request, None) is False

    def test_staff(self, permission_obj, allowed_methods, forbidden_methods):
        self.request.user = self.staff
        for method in allowed_methods:
            self.request.method = method
            assert permission_obj.has_permission(self.request, None) is False

    # pylint: enable=unused-argument


class TestIsStaff:
    @pytest.fixture(autouse=True)
    def before_each(self, user_factory):
        # pylint: disable=attribute-defined-outside-init
        self.factory = RequestFactory()
        self.user = user_factory(basic=True)
        self.staff = user_factory(staff=True)
        self.staff.is_staff = True
        self.permission_obj = identities_permission.IsStaff()
        # pylint: enable=attribute-defined-outside-init

    def test_anonymous(self):
        request = self.factory.request()
        request.user = AnonymousUser()
        assert self.permission_obj.has_permission(request, None) is False

    def test_user(self):
        request = self.factory.request()
        request.user = self.user
        assert self.permission_obj.has_permission(request, None) is False

    def test_staff(self):
        request = self.factory.request()
        request.user = self.staff
        assert self.permission_obj.has_permission(request, None) is True


class TestIsStaffOrReadOnly:
    @pytest.fixture(autouse=True)
    def before_each(self, user_factory):
        # pylint: disable=attribute-defined-outside-init
        self.factory = RequestFactory()
        self.user = user_factory(basic=True)
        self.staff = user_factory(staff=True)
        self.permission_obj = identities_permission.IsStaffOrReadOnly()
        self.request = self.factory.request()
        # pylint: enable=attribute-defined-outside-init

    def test_anonymous(self):
        self.request.user = AnonymousUser()
        assert self.permission_obj.has_permission(self.request, None) is False

    def test_user(self):
        self.request.user = self.user
        for method in permissions.SAFE_METHODS:
            self.request.method = method
            assert self.permission_obj.has_permission(self.request, None) is True
        for method in NONSAFE_METHODS:
            self.request.method = method
            assert self.permission_obj.has_permission(self.request, None) is False

    def test_staff(self):
        self.request.user = self.staff
        for method in permissions.SAFE_METHODS + NONSAFE_METHODS:
            self.request.method = method
            assert self.permission_obj.has_permission(self.request, None) is True


@pytest.mark.parametrize(
    "test_input, expected",
    (
        (permissions.SAFE_METHODS, True),
        (NONSAFE_METHODS, False),
    ),
)
def test_readonly(rf, test_input, expected, basic_user_auth):
    permission_obj = identities_permission.ReadOnly()
    request = rf.request()
    request.user = basic_user_auth
    for method in test_input:
        request.method = method
        assert permission_obj.has_permission(request, None) is expected


class TestObjectPermission:
    @pytest.mark.parametrize(
        "method", ("GET", "OPTIONS", "HEAD", "PUT", "PATCH", "DELETE")
    )
    def test_get_required_object_permissions(self, method):
        ObjectPermission().get_required_object_permissions(method, Profile)

    @pytest.mark.parametrize("method", ("POST", "FOO", "get"))
    def test_get_required_object_permissions_not_allowed_method(self, method):
        with pytest.raises(MethodNotAllowed):
            ObjectPermission().get_required_object_permissions(method, Profile)

    def test_has_permission_anonymous(self):
        request = factory.get("/")
        request.user = None
        assert ObjectPermission().has_permission(request, None) is False

    def test_has_permission_method_not_allowed(self, user_factory):
        request = factory.post("/")
        request.user = user_factory()
        with pytest.raises(MethodNotAllowed):
            ObjectPermission().has_permission(request, None)

    @pytest.mark.parametrize(
        "method", ("get", "options", "head", "put", "patch", "delete")
    )
    def test_has_permission_method_ok(self, user_factory, method):
        request = getattr(factory, method)("/")
        request.user = user_factory()
        assert ObjectPermission().has_permission(request, None) is True

    def test_has_object_permission(self, user_factory):
        request = factory.get("/")
        mock_user = Mock()
        request.user = mock_user
        profile = Profile(user=user_factory())
        ObjectPermission().has_object_permission(request, None, profile)
        mock_user.has_perms.assert_called_once_with(
            (f"{APP_NAME}.view_{profile.__class__._meta.model_name}",), profile
        )



import logging

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from rest_framework import serializers

from ..permissions import has_group_manager_groups
from ..serializers.profile import ProfileSerializer


logger = logging.getLogger(__name__)
User = get_user_model()


class UserinfoSerializer(serializers.ModelSerializer):
    # Sets `required: true` in the OpenAPI schema
    id = serializers.IntegerField(required=True)
    profile = ProfileSerializer(read_only=True)
    ip_address = serializers.SerializerMethodField()
    permissions = serializers.SerializerMethodField()

    class Meta:
        model = User
        read_only_fields = ("ip_address", "permissions")
        fields = read_only_fields + (
            "username",
            "email",
            "first_name",
            "last_name",
            "profile",
            "id",
        )

    def get_ip_address(self, _: AbstractUser):
        request = self.context["request"]
        x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
        if x_forwarded_for:
            return x_forwarded_for.split(",")[0]
        return request.META.get("REMOTE_ADDR")

    def get_permissions(self, _: AbstractUser):
        request = self.context["request"]
        return {
            "staff": request.user.is_staff,
            "group_manager": has_group_manager_groups(request.user),
        }



# pylint: disable=unused-import
from identities.views.group import GroupViewSet
from identities.views.permission import PermissionViewSet, ObjectByPermissionViewSet
from identities.views.user import UserViewSet, UserinfoView



from django.contrib.auth.models import AbstractUser
from django.db import models
from django.dispatch import receiver
from guardian.mixins import GuardianUserMixin
from simple_history.models import HistoricalRecords


class User(AbstractUser, GuardianUserMixin):
    email = models.EmailField(
        blank=True, max_length=254, unique=True, verbose_name="email address"
    )
    history = HistoricalRecords()


def get_anonymous_user_instance(UserModel):
    return UserModel(username="AnonymousUser", email="anonymous_user@localhost")


class Profile(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    affiliation = models.CharField(max_length=512, blank=True)
    affiliation_id = models.CharField(max_length=512, blank=True)

    history = HistoricalRecords()

    def __str__(self):
        return f"Profile ({self.user.username})"

    @property
    def display_id(self):
        return (
            f"ID: {self.user.username.split('@', maxsplit=1)[0]}"
            if "@" in self.user.username
            else None
        )

    @property
    def display_name(self):
        if self.user.first_name or self.user.last_name:
            name = " ".join(filter(bool, (self.user.first_name, self.user.last_name)))
        else:
            name = self.user.username
        display_id = self.display_id and f"({self.display_id})"
        display_email = self.user.email and f"({self.user.email})"
        return " ".join(filter(bool, (f"{name}", display_id, display_email)))


@receiver(models.signals.post_save, sender=User)
def create_or_update_user_profile(
    sender, instance, created, **kwargs
):  # pylint: disable=unused-argument
    if created:
        Profile.objects.create(user=instance)
    instance.profile.save()



from django.contrib.auth.models import Group
from django_drf_utils.views.utils import unique_check
from rest_framework import viewsets

from ..serializers import GroupSerializer
from ..permissions import (
    IsGroupManagerRead,
    IsGroupManagerWrite,
    IsStaff,
    get_group_manager_groups,
)


@unique_check((IsStaff,))
class GroupViewSet(viewsets.ModelViewSet):  # pylint: disable=too-many-ancestors
    serializer_class = GroupSerializer
    queryset = Group.objects.all()
    # pylint: disable=unsupported-binary-operation
    permission_classes = (IsStaff | IsGroupManagerRead | IsGroupManagerWrite,)

    def get_queryset(self):
        requestor = self.request.user
        if requestor.is_staff:
            return Group.objects.all()
        return get_group_manager_groups(requestor)



from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema

from identities.apps import APP_NAME as IDENTITIES_APP_NAME
from identities.permissions import IsGroupManagerRead
from identities.serializers import AnyObjectSerializer, PermissionSerializer
from identities.permissions import IsStaff


class GenericObjectSchema(AutoSchema):
    def get_responses(self, path, method):
        print(path, method)
        responses = super().get_responses(path, method)
        responses["200"]["content"]["application/json"]["schema"] = {
            "type": "array",
            "items": {
                "$ref": responses["200"]["content"]["application/json"]["schema"][
                    "$ref"
                ]
            },
        }
        return responses


class ObjectByPermissionSchema(AutoSchema):
    def get_filter_parameters(self, path, method):
        if method == "GET":
            return [
                {
                    "name": "perm_id",
                    "in": "query",
                    "required": True,
                    "description": (
                        "Permission ids for which available "
                        "objects should be returned"
                    ),
                    "schema": {"type": "array", "items": {"type": "integer"}},
                },
            ]
        return []


# pylint: disable=too-many-ancestors
class PermissionViewSet(viewsets.ReadOnlyModelViewSet):
    # pylint: disable=unsupported-binary-operation
    permission_classes = (IsStaff | IsGroupManagerRead,)

    def get_serializer_class(self):
        if self.action == "objects":
            return AnyObjectSerializer
        return PermissionSerializer

    def get_queryset(self):
        content_types = [
            x.id
            for x in ContentType.objects.filter(
                app_label__in=(IDENTITIES_APP_NAME, "auth")
            ).exclude(model__startswith="historical")
        ]
        return Permission.objects.filter(content_type__in=content_types)

    @action(detail=True, schema=GenericObjectSchema())
    def objects(self, request, pk=None):  # pylint: disable=unused-argument
        return Response(
            AnyObjectSerializer(
                self.get_object().content_type.get_all_objects_for_this_type(),
                many=True,
            ).data
        )


class ObjectByPermissionViewSet(viewsets.ViewSet):
    # pylint: disable=unsupported-binary-operation
    permission_classes = (IsStaff | IsGroupManagerRead,)
    schema = ObjectByPermissionSchema()

    def list(self, request):
        permissions = Permission.objects.filter(
            pk__in=request.query_params.getlist("perm_id")
        )
        return Response(
            {
                p.id: AnyObjectSerializer(
                    p.content_type.get_all_objects_for_this_type(),
                    many=True,
                ).data
                for p in permissions
            }
        )



# Generated by Django 4.0.6 on 2022-11-04 07:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("identities", "0004_alter_historicaluser_id"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalprofile",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical profile",
                "verbose_name_plural": "historical profiles",
            },
        ),
        migrations.AlterModelOptions(
            name="historicaluser",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical user",
                "verbose_name_plural": "historical users",
            },
        ),
        migrations.AlterField(
            model_name="historicalprofile",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicaluser",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]



import factory
from django.contrib.auth import get_user_model
from django.db.models import signals
from django.contrib.auth.models import Group
from identities.models import (
    Profile,
)


USER_PASSWORD = "pass"  # nosec


class GroupFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Group

    name = factory.Sequence(lambda n: f"Group {n}")


@factory.django.mute_signals(signals.post_save)
class ProfileFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Profile

    affiliation = "member@foo.edu,member@bar.org,staff@bar.org"


@factory.django.mute_signals(signals.post_save)
class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = get_user_model()

    username = factory.Sequence(lambda n: f"user_{n}")
    password = factory.PostGenerationMethodCall("set_password", USER_PASSWORD)
    email = factory.Faker("email")
    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    profile = factory.RelatedFactory(ProfileFactory, "user")

    class Params:
        basic = factory.Trait(
            username="user-basic",
            email="foo@bar.org",
            password=factory.PostGenerationMethodCall("set_password", USER_PASSWORD),
        )
        staff = factory.Trait(
            username="user-staff",
            is_staff=True,
            password=factory.PostGenerationMethodCall("set_password", USER_PASSWORD),
        )



import logging

from rest_framework import serializers

from ..models import Profile

logger = logging.getLogger(__name__)


class ProfileSerializer(serializers.ModelSerializer):
    affiliation_home = serializers.SerializerMethodField()

    class Meta:
        model = Profile
        read_only_fields = ("affiliation", "affiliation_id", "affiliation_home")
        fields = read_only_fields + (
            "display_name",
            "display_id",
        )

    def get_affiliation_home(self, obj: Profile):
        if obj.affiliation:
            return ",".join(
                sorted(
                    {aff.split("@")[1].strip() for aff in obj.affiliation.split(",")}
                )
            )
        return ""



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin

from identities.models import (
    Profile,
)

User = get_user_model()


class ProfileInline(admin.StackedInline):
    model = Profile
    can_delete = False
    verbose_name_plural = "Profile"


@admin.register(User)
class UserAdmin(BaseUserAdmin):
    inlines = (ProfileInline,)



import pytest
from django.contrib.auth import get_user_model

from identities.models import Profile

User = get_user_model()


class TestProfile:
    test_input_expected = (
        (
            ("cnorris", "Chuck", "Norris", "cnorris@roundhouse.gov"),
            "Chuck Norris (cnorris@roundhouse.gov)",
        ),
        (
            (
                "CNORRIS@roundhouse.gov",
                None,
                None,
                "cnorris@roundhouse.gov",
            ),
            "CNORRIS@roundhouse.gov (ID: CNORRIS) (cnorris@roundhouse.gov)",
        ),
        (
            ("CNORRIS@roundhouse.gov", None, None, None),
            "CNORRIS@roundhouse.gov (ID: CNORRIS)",
        ),
    )

    @pytest.mark.parametrize("test_input, expected", test_input_expected)
    def test_display(self, test_input, expected):
        profile = Profile(
            user=User(
                username=test_input[0],
                first_name=test_input[1],
                last_name=test_input[2],
                email=test_input[3],
            ),
        )
        assert profile.display_name == expected

    def test_str(self):
        assert str(Profile(user=User(username="bob"))) == "Profile (bob)"
